/*
 * Copyright (c) 2004-2006 Erez Zadok,
 * Copyright (c) 2004-2006 Nikolai Joukov,
 * Copyright (c) 2005-2006 Chaitanya Patti,
 * Copyright (c) 2004-2006 Stony Brook University
 */
/*
 *  $Id: ECC.h,v 1.11 2006/12/13 17:00:21 kolya Exp $
 */

#ifdef __KERNEL__
#ifndef __ECC_H_
#define __ECC_H_

//#include "fec/fec_linux.h"

/*
 * Recover a missing page from parity.
 * No performance optimizations yet.
 */
static inline void recover_ECC_XOR(page_t** pages, int npages, int which)
{
	int i, j;
	char* ECC_page = (char*)(page_address(pages[npages - 1]));
	char* lost_page = (char*)(page_address(pages[which]));

	for (i = 0; i < PAGE_SIZE; i++) {
		lost_page[i] = ECC_page[i];
		for (j = 0; j < npages - 1; j++)
			if (j != which)
				lost_page[i] = lost_page[i] ^ ((char*)(page_address(pages[j])))[i];
	}
}

/*Recover page at (row,col) of RDP array by using diagonal parity*/
static inline void recover_from_diag_RDP(page_t *** page_array, int row, int col, page_t ** par, int bnum, int sus)
{
	int s = 0;

	for(s = 0; s < sus; s++){
		int count = 1, j = 0;
		par[0] = page_array[row * sus + s][col];
		for(j = 0;j < bnum; j++){
			int ncol = j;
			int nrow = (row + col + bnum - 1 - ncol) % (bnum - 1);
				if(nrow != (bnum - 2) && ncol != col){
					if(page_array[nrow * sus + s][ncol]){
						par[count] = page_array[nrow * sus + s][ncol];
						count++;
					}
				}
		}
		if(count >= 2)
			recover_ECC_XOR(par, count, 0);
	}
}

/*Recover page at (row,col) of RDP array by using row parity*/
static inline void recover_from_row_RDP(page_t *** page_array, int row, int col, page_t ** par, int bnum, int sus)
{
	int s = 0;

	for(s = 0; s < sus; s++){
		int count = 1, j = 0;
		par[0] = page_array[row * sus + s][col];
		for(j = 0; j < bnum - 1; j++){
			int ncol = j;
			if((ncol != col) && page_array[row * sus + s][ncol]){
				par[count]=page_array[row * sus + s][ncol];
				count++;
			}
		}
		if(count >= 2)
			recover_ECC_XOR(par, count, 0);
	}
}

/* Recover data in Row-Diagonal Parity(RDP) array. The failed branches are branch1 and branch2.
 * They are reconstructed. We can make this more efficient by reconstructing only the page(s) that are 
 * needed by the calling recovery function*/

static inline int recover_ECC_RDP(page_t *** page_array, int bnum, int branch1, int branch2, int sus, int req_page)
{
	int err = 0;
	int i = 0,j = 0,d;
	page_t ** par = NULL;
	page_t * temp = NULL;

	par = KMALLOC(sizeof(page_t *) * bnum, GFP_KERNEL);

	if(!par){
		err = -ENOMEM;
		goto out;
	}
		
	for(i = 0; i < (bnum - 2) * sus; i++){
		temp = alloc_pages(GFP_KERNEL|__GFP_REPEAT|__GFP_ZERO, 0);
		if(IS_ERR(temp)){
			err = -ENOMEM;
			goto out_clear_array;
		}
		page_array[i][branch1] = temp;

		if(branch2 != -1){
			temp = alloc_pages(GFP_KERNEL|__GFP_REPEAT|__GFP_ZERO, 0);
			if(IS_ERR(temp)){
				err = -ENOMEM;
				goto out_clear_array;
			}
			page_array[i][branch2] = temp;
		}
	}

	if((branch2 == -1) || (branch2 == bnum - 1)){

		for(i = 0;i < (bnum - 2); i++)
			recover_from_row_RDP(page_array, i, branch1, par, bnum, sus);

		if(branch2 == bnum - 1){
			for(d = 0; d < (bnum - 2); d++)
				recover_from_diag_RDP(page_array, d, bnum-1, par, bnum, sus);	
		}

	}
	else{
		int g1, g2;
		
		g1 = (branch1 + bnum - 2) % (bnum - 1);

		g2 = (branch2 + bnum - 2) % (bnum - 1);

		j = (g2 + bnum - 1 - g1) % (bnum - 1);

		while(g1 != (bnum - 2))
		{
			recover_from_diag_RDP(page_array, (g1 + bnum - 1 - branch2) % (bnum - 1), branch2, par, bnum, sus);
			recover_from_row_RDP(page_array, (g1 + bnum - 1 - branch2) % (bnum - 1), branch1, par, bnum, sus);

			if(req_page/sus == (g1 + bnum - 1 - branch2) % (bnum - 1))
				goto out;

			g1 = (g1 + bnum - 1 - j)%(bnum - 1);
		}

		while(g2 != (bnum - 2))
		{
			recover_from_diag_RDP(page_array,(g2 + bnum - 1 - branch1) % (bnum - 1), branch1, par, bnum, sus);

			if(req_page/sus == (g2 + bnum - 1 - branch1) % (bnum - 1))
				goto out;

			recover_from_row_RDP(page_array, (g2 + bnum - 1 - branch1) % (bnum - 1), branch2, par, bnum, sus);
			g2 = (g2 + j) % (bnum - 1);
		}

	}

	goto out;

out_clear_array:

	if(branch1 != -1)
	{
		for(i = 0; i < (bnum - 2) * sus; i++){
			if(page_array[i][branch1]){
				free_page((unsigned long)page_array[i][branch1]);
			}
		}	
	}

	if(branch2 != -1)
	{
		for(i = 0;i < (bnum-2) * sus; i++){
			if(page_array[i][branch2]){
				free_page((unsigned long)page_array[i][branch2]);
			}
		}	
	}
out:
	if(par)
		KFREE(par);
	return err;
}


static inline void update_ECC_XOR_buf(char *oldd, char *xord, char *newd, int from, int to, int fresh)
{
	int i, bytesb, words, bytese;
	unsigned int* ioldd;
	unsigned int* ixord;
	unsigned int* inewd;
	if (to - from < sizeof(unsigned int) + 1) {
		words = 0;
		bytesb = to - from;
		bytese = 0;
	} else {
		bytesb = sizeof(unsigned int) - (from & (sizeof(unsigned int) - 1));
		bytese = to & (sizeof(unsigned int) - 1);
		words = (to - from - bytesb - bytese) / sizeof(unsigned int);
	}

	if (fresh) {
		for (i = from; i < from + bytesb; i++) {
			xord[i] = xord[i] ^ newd[i];
			oldd[i] = newd[i];
		}
		ioldd = (unsigned int *)&oldd[i];
		inewd = (unsigned int *)&newd[i];
		ixord = (unsigned int *)&xord[i];
		for (i = 0; i < words; i++) {
			ixord[i] = ixord[i] ^ inewd[i];
			ioldd[i] = inewd[i];
		}
		for (i = to - bytese; i < to; i++) {
			xord[i] = xord[i] ^ newd[i];
			oldd[i] = newd[i];
		}
	} else {
		for (i = from; i < from + bytesb; i++) {
			xord[i] = xord[i] ^ oldd[i] ^ newd[i];
			oldd[i] = newd[i];
		}
		ioldd = (unsigned int *)&oldd[i];
		inewd = (unsigned int *)&newd[i];
		ixord = (unsigned int *)&xord[i];
		for (i = 0; i < words; i++) {
			ixord[i] = ixord[i] ^ ioldd[i] ^ inewd[i];
			ioldd[i] = inewd[i];
		}
		for (i = to - bytese; i < to; i++) {
			xord[i] = xord[i] ^ oldd[i] ^ newd[i];
			oldd[i] = newd[i];
		}
	}
}


/*
 * Calculate parity in units of words for performance.
 */
static inline void update_ECC_XOR(page_t* old_page_data, char* newd, page_t* page_ECC, int from, int to, int fresh)
{
	char* oldd;
	char* xord;

	if (fresh)
		fist_dprint(8, "     %s: Using fast XOR ECC\n", __FUNCTION__);
	else
		fist_dprint(8, "     %s: Using slow XOR ECC\n", __FUNCTION__);

	oldd = (char*)(page_address(old_page_data));
	xord = (char*)(page_address(page_ECC));

	return update_ECC_XOR_buf(oldd, xord, newd, from, to, fresh);

}


/* Update for the Row-Diagonal Parity Algorithm
 * There are three pages that we need to update:
 * 1. The page containing the row-parity of the row parity set to which this page belongs to.
 * 2. The page containing the diagonal parity of the diagonal parity set to which this page belongs.
 * 3. The page containing the diagonal parity of the diagonal parity set to which the page mentioned in 1 above belongs
 */

static inline void update_ECC_RDP(page_t** pages, char* newd, int from, int to, int fresh){
	int i, bytesb, words, bytese;
	char* oldd;
	char* xord_1,*xord_2,*xord_3;
	char tempch;
	unsigned int* ioldd;
	unsigned int* ixord_1;
	unsigned int* ixord_2;
	unsigned int* ixord_3;
	unsigned int* inewd;
	unsigned int tempint;


	if (fresh)
		fist_dprint(8, "     %s: Using fast XOR RDP ECC\n", __FUNCTION__);
	else
		fist_dprint(8, "     %s: Using slow XOR RDP ECC\n", __FUNCTION__);

	oldd = (char*)(page_address(pages[0]));
	xord_1 = (char*)(page_address(pages[1]));
	if(pages[2])
		xord_2 = (char*)(page_address(pages[2]));
	else
		xord_2 = NULL;

	if(pages[3])
		xord_3 = (char*)(page_address(pages[3]));
	else
		xord_3 = NULL;

	if (to - from < sizeof(unsigned int) + 1) {
		words = 0;
		bytesb = to - from;
		bytese = 0;
	} else {
		bytesb = sizeof(unsigned int) - (from & (sizeof(unsigned int) - 1));
		bytese = to & (sizeof(unsigned int) - 1);
		words = (to - from - bytesb - bytese) / sizeof(unsigned int);
	}

	if (fresh) {
		for (i = from; i < from + bytesb; i++) {
			xord_1[i] = xord_1[i] ^ newd[i];
			if(xord_2)
				xord_2[i] = xord_2[i] ^ newd[i];
			if(xord_3)
				xord_3[i] = xord_3[i] ^ xord_1[i];
			oldd[i] = newd[i];
		}
		ioldd = (unsigned int *)&oldd[i];
		inewd = (unsigned int *)&newd[i];
		ixord_1 = (unsigned int *)&xord_1[i];

		if(xord_2)
			ixord_2 = (unsigned int *)&xord_2[i];
		else
			ixord_2 = NULL;

		if(xord_3)
			ixord_3 = (unsigned int *)&xord_3[i];
		else
			ixord_3 = NULL;

		for (i = 0; i < words; i++) {
			ixord_1[i] = ixord_1[i] ^ inewd[i];
			if(ixord_2)
				ixord_2[i] = ixord_2[i] ^ inewd[i];
			if(ixord_3)
				ixord_3[i] = ixord_3[i] ^ ixord_1[i];
			ioldd[i] = inewd[i];
		}
		for (i = to - bytese; i < to; i++) {
			xord_1[i] = xord_1[i] ^ newd[i];
			if(xord_2)
				xord_2[i] = xord_2[i] ^ newd[i];
			if(xord_3)
				xord_3[i] = xord_3[i] ^ xord_1[i];
			oldd[i] = newd[i];
		}
	} else {
		for (i = from; i < from + bytesb; i++) {
			tempch = xord_1[i];
			xord_1[i] = xord_1[i] ^ oldd[i] ^ newd[i];
			if(xord_2)
				xord_2[i] = xord_2[i] ^ oldd[i] ^ newd[i];
			if(xord_3)
				xord_3[i] = xord_3[i] ^ tempch ^ xord_1[i]; 
			oldd[i] = newd[i];
		}
		ioldd = (unsigned int *)&oldd[i];
		inewd = (unsigned int *)&newd[i];
		ixord_1 = (unsigned int *)&xord_1[i];

		if(xord_2)
			ixord_2 = (unsigned int *)&xord_2[i];
		else
			ixord_2 = NULL;
		if(xord_3)
			ixord_3 = (unsigned int *)&xord_3[i];
		else
			ixord_3 = NULL;

		for (i = 0; i < words; i++) {
			tempint = ixord_1[i];
			ixord_1[i] = ixord_1[i] ^ ioldd[i] ^ inewd[i];
			if(ixord_2)
				ixord_2[i] = ixord_2[i] ^ ioldd[i] ^ inewd[i];
			if(ixord_3)
				ixord_3[i] = ixord_3[i] ^ tempint ^ ixord_1[i]; 
			ioldd[i] = inewd[i];
		}
		for (i = to - bytese; i < to; i++) {
			tempch = xord_1[i];
			xord_1[i] = xord_1[i] ^ oldd[i] ^ newd[i];
			if(xord_2)
				xord_2[i] = xord_2[i] ^ oldd[i] ^ newd[i];
			if(xord_3)
				xord_3[i] = xord_3[i] ^ tempch ^ xord_1[i]; 
			oldd[i] = newd[i];
		}
	}
}



/*
 *	update_ECC - updates error correction codes in a set of pages
 *
 *	@pages:		pointer to array of pages to modify
 *	@newdata:	the data to write to the pages
 *	@which:		which page in the pages array to commit newdata to
 *	@from:		the location to start writing
 *	@to:		the location to stop writing
 *	@npages		how many pages are in the pages array
 *	@iECCw:		the width of ECC pages.  How many ECC pages there are
 *	@fresh:		flag, true->new data page; false->page existed before
 *
 *	This function updates ECC codes for page changes in a RAIF filesystem.
 */
static inline void update_ECC(page_t** pages, char* newdata, int which, int from, int to, int npages, int iECCw, int fresh)
{
	switch(iECCw){
		case 0:
			memcpy((char*)page_address(pages[which]) + from,
				newdata + from, to - from);
			break;
		case 1:
			update_ECC_XOR(pages[which], newdata, pages[npages - 1],
					from, to, fresh);
			break;
		case 2:
		case 3:
		case 4:
		case 5:
		case 6:
		case 7:
		case 8:
#ifdef RAIF6_STRIPE_FEC
			// Untested
			if (update_ECC_FEC(pages, newdata, which, from, to, npages, iECCw))
				panic("update_ECC_FEC(): failed\n");
			break;
#endif
		default:
			BUG();
	}
}


/*
 *	recover_ECC - recovers data, from a set of pages, with read failures
 *
 *	@pages:		pointer to array of pages that were read
 *	@npages		how many pages are in the pages array
 *	@bad:		list of bad pages
 *	@nbad:		how many bad pages
 *	@iECCw:		the width of ECC pages.  How many ECC pages there are
 *
 *	@returns:	0 on success. (TODO: Define some error codes?)
 *	This function recovers data pages, based upon failures to read.
 **/
static inline int recover_ECC(page_t** pages, int npages, int *bad, int nbad, int iECCw)
{
	if (nbad > iECCw)
		return -1;

	if (iECCw == 1) {
		recover_ECC_XOR(pages, npages, bad[0]);
		return 0;
	}
#ifdef RAIF6_STRIPE_FEC
	return recover_ECC_FEC(pages, npages, bad, nbad, iECCw);
#else
	return -1;
#endif
}

#endif
#endif

/*
 * Local variables:
 * c-basic-offset: 4
 * End:
 */
