/*
 * Copyright (c) 2005-2006 Erez Zadok,
 * Copyright (c) 2005      David P. Quigley,
 * Copyright (c) 2005-2006 Stony Brook University,
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package, or get one from ftp://ftp.filesystems.org/pub/fist/COPYING.
 *
 * This Copyright notice must be kept intact and distributed with all
 * fistgen sources INCLUDING sources generated by fistgen.
 */
/*
 *  $Id: missing_vfs_funcs.h,v 1.5 2006/12/30 22:08:37 crpatti Exp $
 */

#ifndef __MISSING_VFS_FUNCS_H_
#define __MISSING_VFS_FUNCS_H_
/*
 * DQ: These functions used to be in the actual vfs but now they are
 * missing so we redefine them here if we are using 2.6 this code is
 * used in the 2.6 templates.
 * Common dentry functions for inclusion in the VFS
 * or in other stackable file systems.  Some of these
 * functions were in linux/fs/ C (VFS) files.
 *
 */

/*
 * Locking the parent is needed to:
 *  - serialize directory operations
 *  - make sure the parent doesn't change from
 *    under us in the middle of an operation.
 *
 * NOTE! Right now we'd rather use a "struct inode"
 * for this, but as I expect things to move toward
 * using dentries instead for most things it is
 * probably better to start with the conceptually
 * better interface of relying on a path of dentries.
 */
static inline struct dentry *lock_parent(struct dentry *dentry) {
        struct dentry *dir = dget(dentry->d_parent);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
        down(&dir->d_inode->i_sem);
#else
        mutex_lock(&dir->d_inode->i_mutex);
#endif
        return dir;
}

static inline struct dentry *get_parent(struct dentry *dentry) {
        return dget(dentry->d_parent);
}

static inline void unlock_dir(struct dentry *dir)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
        up(&dir->d_inode->i_sem);
#else
        mutex_unlock(&dir->d_inode->i_mutex);
#endif
        dput(dir);
}

/*
 * Whee.. Deadlock country. Happily there are only two VFS
 * operations that do this..
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
static inline void double_down(struct semaphore *s1, struct semaphore *s2)
#else
static inline void double_down(struct mutex *s1, struct mutex *s2)
#endif
{
        if (s1 != s2) {
                if ((unsigned long) s1 < (unsigned long) s2) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
                        struct semaphore *tmp;
#else
			struct mutex *tmp;
#endif
			tmp = s2;
                        s2 = s1; s1 = tmp;
                }
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
                down(s1);
#else
		mutex_lock(s1);
#endif
        }
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
        down(s2);
#else
	mutex_lock(s2);
#endif
}

/*
 * Ewwwwwwww... _triple_ lock. We are guaranteed that the 3rd argument is
 * not equal to 1st and not equal to 2nd - the first case (target is parent of
 * source) would be already caught, the second is plain impossible (target is
 * its own parent and that case would be caught even earlier). Very messy.
 * I _think_ that it works, but no warranties - please, look it through.
 * Pox on bloody lusers who mandated overwriting rename() for directories...
 */

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
static inline void triple_down(struct semaphore *s1,
                               struct semaphore *s2,
                               struct semaphore *s3)
#else
static inline void triple_down(struct mutex *s1,
                               struct mutex *s2,
                               struct mutex *s3)
#endif
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
	struct semaphore *tmp;
#else
	struct mutex *tmp;
#endif
        if (s1 != s2) {
                if ((unsigned long) s1 < (unsigned long) s2) {
                        if ((unsigned long) s1 < (unsigned long) s3) {
                                tmp = s3;
                                s3 = s1; s1 = tmp;
                        }
                        if ((unsigned long) s1 < (unsigned long) s2) {
                                tmp = s2;
                                s2 = s1; s1 = tmp;
                        }
                } else {
                        if ((unsigned long) s1 < (unsigned long) s3) {
                                tmp = s3;
                                s3 = s1; s1 = tmp;
                        }
                        if ((unsigned long) s2 < (unsigned long) s3) {
                                tmp = s3;
                                s3 = s2; s2 = tmp;
                        }
                }
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
                down(s1);
#else
		mutex_lock(s1);
#endif
        } else if ((unsigned long) s2 < (unsigned long) s3) {
                tmp = s3;
                s3 = s2; s2 = tmp;
        }
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
        down(s2);
        down(s3);
#else
	mutex_lock(s2);
	mutex_lock(s3);
#endif
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
static inline void double_up(struct semaphore *s1, struct semaphore *s2)
#else
static inline void double_up(struct mutex *s1, struct mutex *s2)
#endif
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
        up(s1);
#else
	mutex_unlock(s1);
#endif
        if (s1 != s2) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
                up(s2);
#else
		mutex_unlock(s2);
#endif
	}
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
static inline void triple_up(struct semaphore *s1,
                             struct semaphore *s2,
                             struct semaphore *s3)
#else
static inline void triple_up(struct mutex *s1,
                             struct mutex *s2,
                             struct mutex *s3)
#endif
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
        up(s1);
#else
	mutex_unlock(s1);
#endif
        if (s1 != s2) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
                up(s2);
#else
		mutex_unlock(s2);
#endif
	}
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
        up(s3);
#else
	mutex_unlock(s3);
#endif
}

static inline void double_lock(struct dentry *d1, struct dentry *d2)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
        double_down(&d1->d_inode->i_sem, &d2->d_inode->i_sem);
#else
        double_down(&d1->d_inode->i_mutex, &d2->d_inode->i_mutex);
#endif
}

static inline void double_unlock(struct dentry *d1, struct dentry *d2)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
        double_up(&d1->d_inode->i_sem,&d2->d_inode->i_sem);
#else
        double_up(&d1->d_inode->i_mutex,&d2->d_inode->i_mutex);
#endif
        dput(d1);
        dput(d2);
}

#endif //__MISSING_VFS_FUNCS_H_

